#if !defined __ALPHABET_H__
#define __ALPHABET_H__

#include <string>
#include <vector>

#include "checks.h"
#include "classError.h"


//class std::istream;
//class std::ostream;

class Alphabet {

  typedef std::string string;
  typedef std::vector<string> Rep;
  typedef Rep::iterator iterator;

public:
  typedef Rep::const_iterator const_iterator;
  typedef Rep::size_type size_type;

private:
  iterator begin() { return data.begin();}
  iterator end() { return data.end();}

  string &operator[](int i) {
    PRECONDITION(i>=0 && i<int(data.size()));
    return data[i];
  }

public:


  const_iterator begin() const { return data.begin();}
  const_iterator end() const { return data.end();}
  size_type size() const {return data.size();}

  Alphabet() {}
  Alphabet(const Alphabet &abc):data(abc.data) {}
  Alphabet(const char *str) {
    assign(str);
  }
  template<class Iter> Alphabet(Iter a,Iter b) {
    assign(a,b);
  }

  Alphabet &operator=(const Alphabet &abc) {
    data = abc.data;
    return *this;
  }

  const string &operator[](int i) const {
    PRECONDITION(i>=0 && i<int(data.size()));
    return data[i];    
  }

  int at(const string &vname) const;

  void assign(const char *s);
  template<class Iter> void assign(Iter a,Iter b);

  std::ostream &print(std::ostream &os) const;
  std::istream &scan(std::istream &is);

private:
  Rep data;
};

typedef ClassError<Alphabet> AlphabetError;

template<class Iter> void Alphabet::assign(Iter a,Iter b) {
  data.clear();
  for(Iter i=a;i!=b;++i) {
    if ( find(data.begin(),data.end(),*i)!=data.end() ) throw AlphabetError("duplicated var");
    data.push_back(*i);
  }
  if (data.empty()) throw AlphabetError("empty alphabet assignment");
}

inline std::ostream &operator<<(std::ostream &os, const Alphabet &abc) {
  return abc.print(os);
}

inline std::istream &operator>>(std::istream &is, Alphabet &abc) {
  return abc.scan(is);
}

#endif //__ALPHABET_H__
