#include <cstring>
#include "checks.h"
#include "baseparser.h"


using namespace std;


ostream &SlocString::print(std::ostream &os) const {
	return os<<(m_indeed ? "YES":"NO")<<' '<<locString();
}

inline SlocString alt(const SlocString &s1,const SlocString &s2) {
	SlocString p1=s1;
	return (bool)p1 ? p1:s2;
}

inline SlocString seq(const SlocString &s1,const SlocString &s2) {
	SlocString p1=s1;\
	if ( (bool)p1 ) {SlocString p2=s2;p1=SlocString( (bool)p2,p1+p2);};
	return p1;
}


inline SlocString opt1st(const SlocString &s1,const SlocString &s2) {
	SlocString p1=s1;
	SlocString p2=s2;
	if ((bool)p1 && (bool)p2) p2=SlocString(true,p1+p2);
	return p2;
}

inline SlocString opt2nd(const SlocString &s1,const SlocString &s2) {
	SlocString p1=s1;
	if ( (bool)p1 ) {SlocString p2=s2; if ( (bool)p2 ) p1=SlocString(true,p1+p2);};
	return p1;
}

void QueueScanner::clearQueue() {
	while (!m_queue.empty()) m_queue.pop();
}

Token QueueScanner::getToken() {
	if (m_queue.empty()) return scanner.get();
	else {
		Token token=m_queue.front();
		m_queue.pop();
		return token;
	}
}

void QueueScanner::ungetToken(const Token &token) {
	m_queue.push(token);
}

char ParserBase::selSymb(const SlocString &ls) {
	if (!(bool)ls || ls.data().size()!=1) return 0;
	return *ls.data().begin();
}

const char *ParserBase::validSymb(const LocString &ls,const char *valid) {
	PRECONDITION(ls.data().size()==1);
	return strchr(valid,*ls.data().begin());
}

SlocString ParserBase::symb(const char *valid) {
	Token token=getToken();
	if ( token.type()==Token::Symbol && validSymb(token,valid) ) return SlocString(true,token);
	ungetToken(token);
	return SlocString(false,token);
}

SlocString ParserBase::E() {
	Token token=getToken();
	if ( token.type()==Token::Name && (token.data()=="E" || token.data()=="e") ) return SlocString(true,token);
	ungetToken(token);
	return SlocString(false,token);
}

SlocString ParserBase::end() {
	Token token=getToken();
	if (token.type()==Token::End) return SlocString(true,token);
	ungetToken(token);
	return SlocString(false,token);
}

SlocString ParserBase::name() {
	Token token=getToken();
	if (token.type()==Token::Name) return SlocString(true,token);
	ungetToken(token);
	return SlocString(false,token);
}

SlocString ParserBase::digits() {
	Token token=getToken();
	if (token.type()==Token::Digits) return SlocString(true,token);
	ungetToken(token);
	return SlocString(false,token);
}

SlocString ParserBase::uNumber() {
	return digits();
}

SlocString ParserBase::iNumber() {
	SlocString p1=symb("+-");
	SlocString p2=uNumber();
	if ((bool)p1 && (bool)p2) p2=SlocString(true,p1+p2);
	return p2;
}


SlocString ParserBase::fNumber() {
	SlocString p1=iNumber();
	if (p1) {
		SlocString p2=symb(".");
		if (p2) p1+=p2;
		p2=uNumber();
		if (p2) p1+=p2;
	}
	else {
		p1=symb(".");
		if (!p1) return SlocString(false,p1);
		SlocString p2=uNumber();
		p1+=p2;
		if (!p2) return SlocString(false,p1);
	}
	SlocString p2=E();
	if (p2) {
		p1+=p2;
		SlocString p3=iNumber();
		p1+=p3;
		if (!p3) return SlocString(false,p1);
	}
	return p1;
}
