#include <algorithm>
#include <iomanip>
#include "borderbasis.h"
#include "alt_binary_compose.h"
#include "ematrix.h"
#include "string_cast.h"

using namespace std;


bool BorderBasis::close(const Term &s,const SetOfTerms &ns, SetOfTerms &is, BorderBasis &bb, Pcounter &bcount) const {
	if (ns.contains(s)) {
		is.insert(s);
		return false;
	}
	else {
		const_iterator r=find_if(begin(),end(),compose1(bind2nd(mem_fun_ref(&Term::divides),s),__gnu_cxx::select1st<Bpolin>()));
		PRECONDITION(r!=end());
		const_iterator q=r;
		for (;;) {
			++r;
			r=find_if(r,end(),compose1(bind2nd(mem_fun_ref(&Term::divides),s),__gnu_cxx::select1st<Bpolin>()));
			if (r==end()) break;
			if (lm(*q).coeff().abs() < lm(*r).coeff().abs()) q=r;
		}
		if (s==q->first) {
			bb.add(s,q->second);
		}
		else {
			bb.add(s,Tpolin(q->second*Eterm(s.divide(q->first)),bcount));
		}
		return true;
	}
}

BorderBasis BorderBasis::close(int i,const SetOfTerms &ns, SetOfTerms &is) const {
	is.clear();
	BorderBasis bb;
	Pcounter bcount('B');
	for (SetOfTerms::const_iterator t=ns.begin();t!=ns.end();++t) {
		Term s=*t*t->var(i);
		close(s,ns,is,bb,bcount);
	}
	return bb;
}

BorderBasis BorderBasis::close(const SetOfTerms &ns) const {
	BorderBasis bb;
	if (ns.empty()) return bb;
	Pcounter bcount('B');
	SetOfTerms is;
	size_t n=ns.begin()->context().abc().size();
	for (SetOfTerms::const_iterator t=ns.begin();t!=ns.end();++t) {
		for (size_t i=0;i<n;++i) {
			Term s=*t*t->var(i);
			if (is.find(s)!=is.end() || bb.find(s)!=bb.end()) continue; // it is possible that the same monom is attained several times
			close(s,ns,is,bb,bcount);
		}
	}
	return bb;
}

void BorderBasis::check(const SetOfTerms &ns, const SetOfTerms &bs,Report &r) const {
	const char *man="INTERNAL ERROR: Border basis invariant failed: ";
	ostream &rs=r.rs();
	for (const_iterator i=begin();i!=end();++i) {
		if (!bs.contains(i->first)) {
			rs<<man<<"the (formal) leading term "<<i->first<<" of "<<i->second.tag()<<" is not in the border set\n";
		}
		for (Polin::const_iterator m=i->second.begin();m!=i->second.end();++m) {
			if (m->eterm().main()==i->first) continue;
			if ( !ns.contains(m->eterm().main()) ) {
				rs<<man<<" the inner term "<<m->eterm().main()<<" of "<<i->second.tag()<<" is not in the normal set\n";
			}
		}
	}
}

void BorderBasis::print(std::ostream &os) const {
	if (empty()) {
		os<<"<empty>\n";
		return;
	}
	int width=string_cast( max_element(begin(),end(),
																							alt_compose2(
																										less<string::size_type>(),
																										compose1(
																												mem_fun_ref<string::size_type,string>(&string::size),
																												compose1(	
																														string_caster<Term>(),
																														__gnu_cxx::select1st<value_type>()
																												)
																										)
																							)
												       			)->first
												).size();

	for(const_iterator i=begin();i!=end();++i) {
  	os<<setw(width)<<string_cast(i->first)<<": "<<i->second<<'\n';
	}
}

