#include "cset.h"

using namespace std;

void Cset::print(ostream &os,const Eshift &es) const {
	const Tpolin &bp=m_table[es.first];
	os<<"Eshift("<<bp.tag()<<','<<es.second.tlt()<<','<<es.second.ediff()<<')';
}

void Cset::print(ostream &os,const Eorbit &eorb) const {
	if (eorb.empty()) {
		os<<"<empty E-orbit>\n";
		return;
	}
	Eorbit::pq_reverse_iterator t=eorb.pq_rbegin();
	Eorbit::pq_reverse_iterator first=t;
	Eorbit::pq_reverse_iterator last=eorb.pq_rend();
	for (;first!=last;++first) {
		if ( !eorb.tie(**t,**first) ) {
			os<<"-----------------------\n";
			t=first;
		}
		print(os,**first);os<<'\n';
	}
}

void Cset::print(ostream &os,const Spairs &s) const {
	if (s.empty()) {
		os<<"<empty Spairs list>\n";
		return;
	}
	Spairs::pq_reverse_iterator t=s.pq_rbegin();
	Spairs::pq_reverse_iterator first=t;
	Spairs::pq_reverse_iterator last=s.pq_rend();
	for (;first!=last;++first) {
		if ( !s.tie(**t,**first) ) {
			os<<"-----------------------\n";
			t=first;
		}
		print(os,**first);os<<'\n';
	}
}

void Cset::print(ostream &os,const Spair &sp) const {
	const Tpolin &bp1=m_table[sp.first.first];
	const Tpolin &bp2=m_table[sp.first.second];
	const Eterm &lcm=sp.second.lcm();
	os<<"Spair("<<bp1.tag()<<','<<bp2.tag()<<','<<lcm<<','<<lcm.edegree()<<','<<lcm.tdegree()<<')';	
}

void Cset::printBrief(ostream &os,const Eshift &es) const {
	os<<es.second.ediff();
}

void Cset::printBrief(ostream &os,const Spair &sp) const {
	os<<sp.second.lcm().edegree();
}

void Cset::printBrief(ostream &os,const Eorbit &eorb) const {
	os<<'[';
	if ( !eorb.empty() ) {
		Eorbit::pq_reverse_iterator t=eorb.pq_rbegin();
		Eorbit::pq_reverse_iterator first=t;
		Eorbit::pq_reverse_iterator last=eorb.pq_rend();
		printBrief(os,**first);
		while (++first!=last) {
			if (!eorb.tie(**t,**first)) {
				os<<'|';
				t=first;
			}
			else {
				if ((**first).second.ediff()>9) os<<',';
			}
			printBrief(os,**first);
		}
	}
	os<<']';
}

void Cset::printBrief(ostream &os,const Spairs &s) const {
	os<<'[';
	if ( !s.empty() ) {
		Spairs::pq_reverse_iterator t=s.pq_rbegin();
		Spairs::pq_reverse_iterator first=t;
		Spairs::pq_reverse_iterator last=s.pq_rend();
		printBrief(os,**first);
		while (++first!=last) {
			if (!s.tie(**t,**first)) {
				os<<'|';
				t=first;
			}
			else {
				if ((**first).second.lcm().edegree()>9) os<<',';
			}
			printBrief(os,**first);
		}
	}
	os<<']';
}

void Cset::print(ostream &os) const {
	os<<"S-pairs:\n";
	print(os,m_spairs);
	os<<"Done S-pairs:\n";
	print(os,m_sdone);
	os<<"E-orbit:\n";
	print(os,m_eorbit);
	os<<"Done E-orbit:\n";
	print(os,m_edone);
}

void Cset::printBrief(ostream &os) const {
	os<<"S-pairs: ";
	printBrief(os,m_spairs);os<<'\n';
	os<<"E-orbit: ";
	printBrief(os,m_eorbit);os<<'\n';
}

Tpolin Cset::eval(const Eshift &eshift) const {
	int k=eshift.first;
	const Tpolin bp=m_table[k];
	const PolyContext &ctx=eshift.second.tlt().context();
	const Eterm ep( ctx, eshift.second.ediff() );
	return Tpolin(bp*ep,m_ecounter);
}

pair<Tpolin,Tpolin> Cset::eval(const Spair &spair) const {
	int i=spair.first.first;
	int j=spair.first.second;
	const Eterm &lcm=spair.second.lcm();
	const Tpolin bp1=m_table[i];
	const Tpolin bp2=m_table[j];
	const Eterm s1=lcm.divide(bp1.lt());
	const Eterm s2=lcm.divide(bp2.lt());
	const Tpolin r1=Tpolin(bp1*s1,m_scounter);
	const Tpolin r2=Tpolin(bp2*s2,m_scounter);
	return pair<Tpolin,Tpolin>( r1,r2 );
}

pair<int,int> Cset::add(const Tpolin &bp) {
	int k=m_table.size();
	m_table.push_back(bp);
	pair<int,int> bb1hits(0,0);
	if ( bp.lt().edegree()>0 ) {
		m_eorbit.insert(k,bp);
		REPORT(3,"Inserted e-shift of "<<bp.tag()<<".\n");
	}
	else {
		m_edone.insert(k,bp);
		REPORT(3,"The e-shift of "<<bp.tag()<<" is excluded by coprime.\n");
		++bb1hits.first;
	}
	for (int i=0;i<k;++i) {
		const Tpolin &p1=m_table[i];
		if ( p1.lt().isCoprime(bp.lt()) ) {
			m_sdone.insert(i,k,m_table[i],bp);
			REPORT(3,"The s-pair ("<<p1.tag()<<','<<bp.tag()<<") is excluded by coprime.\n");
			++bb1hits.second;
		}
		else {
			m_spairs.insert(i,k,p1,bp);
			REPORT(3,"Inserted s-pair ("<<p1.tag()<<','<<bp.tag()<<").\n");
		}
	}
	return bb1hits;
}

void Cset::add(const Basis &b) {
	REPORT(2,"Forming critical elements from \n"<<b<<'\n');
	REPORT(1,"The current critical set:\n";printBrief(rs()));
	REPORT(4,"which is:\n"<<*this);
	int shed=m_table.size();
	pair<int,int> bb1hits(0,0);
	for(Basis::const_iterator bi=b.begin();bi!=b.end();++bi) {
		pair<int,int> bb1=add(*bi);
		bb1hits.first+=bb1.first;
		bb1hits.second+=bb1.second;
	}
	stat()["bb1es"]+=bb1hits.first;
	stat()["bb1sp"]+=bb1hits.second;
	REPORT(1,bb1hits.first<<" e-shifts and "<<bb1hits.second<<" s-pairs excluded by coprime.\n");
	REPORT(1,"The critical set before chain:\n";printBrief(rs()));
	REPORT(4,"which is:\n"<<*this);
	bb2(shed);
	REPORT(1,"The critical set after chain:\n";printBrief(rs()));
	REPORT(4,"which is:\n"<<*this);
}

void Cset::clear() {
	m_eorbit.clear();
	m_spairs.clear();
	m_edone.clear();
	m_sdone.clear();
	m_table.clear();
}

int Cset::bb2eorb(int shed) {
	int hits=0;
	Eorbit::pq_iterator first=m_eorbit.pq_begin();
	Eorbit::pq_iterator  last=m_eorbit.pq_end();
	while(first!=last) {
		int k=(**first).first;
		bool exclude=false;
		for(int i=(k<shed ? shed:0);i<int(m_table.size());++i) {
			if (i==k) continue;
			if (
					(m_eorbit.find(i)!=m_eorbit.end() || m_edone.find(i)!=m_edone.end()) &&
					(m_spairs.find(spair_key(i,k))!=m_spairs.end() || m_sdone.find(spair_key(i,k))!=m_sdone.end()) &&
					m_table[i].lt().main().divides((**first).second.tlt())
			) {
				exclude=true;
				break;
			}
		}
		if (exclude) {
			REPORT(3,"the ";print(rs(),**first);rs()<<" is excluded by chain.\n");
			m_eorbit.erase(*first++);
			++hits;
		}
		else {
			++first;
		}
	}
	REPORT(1,hits<<" e-shifts excluded by chain.\n");
	stat()["bb2es"]+=hits;
	return hits;
}

int Cset::bb2sp(int shed) {
	int hits=0;
	Spairs::pq_iterator first=m_spairs.pq_begin();
	Spairs::pq_iterator  last=m_spairs.pq_end();
	while(first!=last) {
		spair_key key=(**first).first;
		bool exclude=false;
		for(int i=(key.second<shed ? shed:0);i<int(m_table.size());++i) {
			if (i==key.first || i==key.second) continue;
			if (
					(m_spairs.find(spair_key(i,key.first))!=m_spairs.end() || m_sdone.find(spair_key(i,key.first))!=m_sdone.end()) &&
					(m_spairs.find(spair_key(i,key.second))!=m_spairs.end() || m_sdone.find(spair_key(i,key.second))!=m_sdone.end()) &&
					m_table[i].lt().divides((**first).second.lcm())
			) {
				exclude=true;
				break;
			}
		}
		if (exclude) {
			REPORT(3,"the ";print(rs(),**first);rs()<<" is excluded by chain\n");
			m_spairs.erase(*first++);
			++hits;
		}
		else {
			++first;
		}
	}
	REPORT(1,hits<<" s-pairs excluded by chain.\n");
	stat()["bb2sp"]+=hits;
	return hits;
}

int Cset::bb2(int shed) {
	int r1=bb2eorb(shed);
	int r2=bb2sp(shed);
	return r1+r2;
}

int Cset::deliver(Basis &b) {
	PRECONDITION(!m_spairs.empty() || !m_eorbit.empty());
	int ndel=0;
	if (m_spairs.empty()) {
		Eorbit::pq_reverse_iterator t=m_eorbit.pq_rbegin();
		Eorbit::pq_reverse_iterator first=t;
		Eorbit::pq_reverse_iterator last=m_eorbit.pq_rend();
		for (;first!=last && m_eorbit.tie(**t,**first);++first) {
			Tpolin p=eval(**first);
			REPORT(3,"Evaluated ";print(rs(),**first);rs()<<" is "<<p<<'\n');
			if (!p.empty()) {
				b.push_back(p);
				++ndel;
			}
		}		
		Eorbit::pq_iterator k=t.base();
		for(Eorbit::pq_iterator i=first.base();i!=k;) {
			m_edone.insert(**i);
			m_eorbit.erase(*i++);
		}
		REPORT(2,"Delivered "<<ndel<<"evaluated e-shifts:\n"<<b<<'\n');
		stat()["es"]+=ndel;
		return 1;
	}
	else {
		Spairs::pq_reverse_iterator t=m_spairs.pq_rbegin();
		Spairs::pq_reverse_iterator first=t;
		Spairs::pq_reverse_iterator last=m_spairs.pq_rend();
		for(;first!=last && m_spairs.tie(**t,**first);++first) {
			pair<Tpolin,Tpolin> sp=eval(**first);
			REPORT(3,"Evaluated ";print(rs(),**first);rs()<<" is\n"<<sp.first<<'\n'<<sp.second<<'\n');
			PRECONDITION(!sp.first.empty() && !sp.second.empty());
			b.push_back(sp.first);
			b.push_back(sp.second);
			++ndel;
		}
		Spairs::pq_iterator k=t.base();
		for(Spairs::pq_iterator i=first.base();i!=k;) {
			m_sdone.insert(**i);
			m_spairs.erase(*i++);
		}
		REPORT(2,"Delivered "<<ndel<<" evaluated s-pairs:\n"<<b<<'\n');
		stat()["sp"]+=ndel;
		return 2;
	}
}
