#include <ostream>
#include "printematrix.h"
#include "alt_binary_compose.h"
#include "ematrix.h"

using namespace std;

void Elimination::Cols::add(const SetOfEterms &st) {
	for(basis_iterator p=m_basis.begin();p!=m_basis.end();++p) {
		for (SetOfEterms::const_iterator t=st.begin();t!=st.end();++t) {
			Polin::const_iterator m=p->monom(*t);
			if (m!=p->end()) {
				//cout<<*p<<'('<<*m<<") is inserted into "<<*t<<'\n';
				(*this)[*t].insert(pivot_iterator(p,distance(p->begin(),m)));
			}
		}
	}
	for (SetOfEterms::const_iterator t=st.begin();t!=st.end();++t) (*this)[*t]; //ensure that all requirested terms are indexed in the matrix
}

void Elimination::Cols::insert_cols(basis_iterator bi) {
  Polin::const_iterator m=bi->begin();
  iterator col=end();
  while (col!=begin() && m!=bi->end()) {
    --col;
    if (col->first==m->eterm()) {
      pivot_iterator rec(bi,distance(bi->begin(),m));
      pair<col_iterator,bool> result=col->second.insert(rec);
      PRECONDITION(result.second);
      ++m;
      continue;
    }
    m=std::lower_bound(m,bi->end(),col->first,[](const Monom &m, const Eterm &key) {return m.eterm()>key;});
    if (m!=bi->end()) col=upper_bound(m->eterm());
  }
}

void Elimination::Cols::erase_cols(basis_iterator bi) {
  Polin::const_iterator m=bi->begin();
  iterator col=end();
  while (col!=begin() && m!=bi->end()) {
    --col;
    if (col->first==m->eterm()) {
      pivot_iterator rec(bi,distance(bi->begin(),m));
      size_type result=col->second.erase(rec);
      PRECONDITION(result);
      ++m;
      continue;
    }
    m=std::lower_bound(m,bi->end(),col->first,[](const Monom &m,const Eterm &key) {return m.eterm()>key;});
    if (m!=bi->end()) col=upper_bound(m->eterm());
  }
}

#define REPORTEM(lev,em) if (printLevel()>=lev) {pe(rs(),em,matrWidth());}

void Elimination::eliminate(Ematrix &em, const SetOfEterms &ired, const SetOfEterms &ikeep,  bool keePivots) {
	SetOfRows extra;
	Cols matrix(em);
	int hits=0;
	SetOfEterms red=ired;
	SetOfEterms keep=ikeep;
	PrintEmatrix pe(red);
	matrix.add(red);
	REPORT(1,"Entering elimination.\n");
	REPORT(2,"The initial polynomials\n"<<em);
	REPORT(2,"The initial matrix\n");
	REPORTEM(2,em);
	int nred=red.size();
	int nkeep=keep.size();
	int printFilter=0;
	//matrix.print(cout);
	while (!red.empty()) {
    col_iterator pivot;
		bool havePivot=matrix.pivotR(pivot,red,extra);
		if (!havePivot) {
			REPORT(3,"The following reducible terms are either gone somewhere at the earlier stage or stand for free variables "<<red<<'\n');
/*			if (printLevel()>=3) {
				rs()<<"The pivots: "<<extra<<'\n'<<"The cols index:\n";
				matrix.print(rs());
			}*/
			break;
		}
    const Eterm &rterm=pivot->pm().eterm();
    const Tpolin &q=**pivot;
		REPORT(3,"Eliminating reducible "<<rterm<<" by "<<q.tag()<<'\n');
		REPORT(4,"with "<<q<<'\n');
		pe.swap(*--red.end(),rterm);
		red.erase(rterm);
/*		cout<<"Forbidden pivots: ";
		for (SetOfRows::const_iterator ri=extra.begin();ri!=extra.end();++ri) cout<<" "<<(**ri).tag();
		cout<<'\n';*/
		col_iterator first=matrix.col_begin(rterm);
		col_iterator  last=matrix.col_end(rterm);
		while (first!=last) {
			if (first == pivot ) {
				++first;
				continue;
			}
      const Tpolin &p=**first;
      const Coeff c=-first->pm().coeff()/pivot->pm().coeff();
      REPORT(3,p.tag()<<" := "<<p.tag()<<" + "<<q.tag()<<" * "<<c<<" --->\n");
      REPORT(4,p.tag()<<" := "<<p.polin()<<" +\n      "<<q.polin()<<" *\n      "<<c<<" --->\n");
      Tpolin r(p.eliminate(q,c,first->pindex(),pivot->pindex()),p.tag());
      REPORT(3,r<<" -n->\n");
      r.normalize();
      REPORT(3,r<<" -e->\n");
      r.calibrateEdegree(ired);
      REPORT(3,r<<'\n');
      ++hits;      
      bool inExtra=extra.contains(*first);
      if (inExtra) extra.erase(*first);
#if 0
      matrix.print(rs());
      REPORT(3,"DEBUG: "<<**first);
      rs().flush();
#endif
      matrix.erase(*first++);
			if (!r.empty()) {
				basis_iterator pos=matrix.insert(r);
				if (inExtra) extra.insert(pos);
			}
			REPORTEM(5,em);
		}
		if (printLevel()>0 && !(++printFilter&0x0F))  {
      cout<<"\rRed elimination: "<<(100-red.size()*100/nred)<<"% complete   ";
      cout.flush();
    }
		if (keePivots) {
      extra.insert(*pivot);
			//REPORT(3,"The pivots: "<<extra<<'\n');
    }
		else {
			REPORT(3,"removing pivot at "<<rterm<<'\n');
			REPORT(4,"with "<<q<<'\n');
			matrix.erase(*pivot);
			REPORTEM(5,em);
		}
	}

	matrix.clear();
	matrix.add(keep);
	while (!keep.empty()) {
    col_iterator pivot;
		bool havePivot=matrix.pivotK(pivot,keep);
		if (!havePivot) {
			REPORT(3,"The following terms to keep are gone somewhere at the earlier stage "<<keep<<'\n');
			break;
		}
    const Eterm &kterm=pivot->pm().eterm();
		const Tpolin &q=**pivot;
		keep.erase(kterm);
		REPORT(3,"Eliminating syzygy term "<<kterm<<" by "<<q.tag()<<'\n');
		REPORT(4,"with "<<q<<'\n');
		col_iterator first=matrix.col_begin(kterm);
		col_iterator  last=matrix.col_end(kterm);
		while (first!=last) {
			if (first == pivot) {
				++first;
				continue;
			}
			const Tpolin &p=**first;
			const Coeff c=-first->pm().coeff()/pivot->pm().coeff();
			REPORT(3,p.tag()<<" := "<<p.tag()<<" + "<<q.tag()<<" * "<<c<<" --->\n");
			REPORT(4,p.tag()<<" := "<<p.polin()<<" +\n      "<<q.polin()<<" *\n      "<<c<<" --->\n");
			Tpolin r(p.eliminate(q,c,first->pindex(),pivot->pindex()),p.tag());
      REPORT(3,r<<" -n->\n");
      r.normalize();
      REPORT(3,r<<" -e->\n");
      r.calibrateEdegree(ired);
			REPORT(3,r<<'\n');
			++hits;      
			bool inExtra=extra.contains(*first);
			if (inExtra) extra.erase(*first);
			matrix.erase(*first++);
			if (!r.empty()) {
				basis_iterator pos=matrix.insert(r);
				if (inExtra) extra.insert(pos);
			}
			REPORTEM(5,em);
		}
		if (printLevel()>0 && !(++printFilter&0x0F))  {
      cout<<"\rKeep elimination: "<<(100-keep.size()*100/nkeep)<<"% complete   ";
      cout.flush();
    }
	}
	matrix.clear();
	if (keePivots) { // split the output into the elimination module(?) basis and the simplifiers
		for (SetOfRows::const_iterator s=extra.begin();s!=extra.end();) {
			REPORT(3,"moving "<<(**s).tag()<<" to simplifies\n");
			m_simp.push_back(**s);
			em.erase(*s);
			extra.erase(s++);
		}
		REPORT(2,"The matrix with simplifiers removed\n");
		REPORTEM(2,em);
		REPORT(2,"The set of simplifies\n"<<simp());
	}
	else {
		REPORT(2,"The matrix after elimination\n");
		REPORTEM(2,em);
	}
	stat()["elim"]+=hits;
	REPORT(1,"Leaving with "<<hits<<" eliminations done.\n");
}

void Elimination::operator()(Basis &b, const SetOfEterms &ired, const SetOfEterms &ikeep,  bool keePivots) {
	clear();
	b.clearDivs();
	eliminate(b,ired,ikeep,keePivots);
	b.reDivs();
}
