#if !defined(__ETERM_H__)
#define __ETERM_H__

#include "term.h"

class Eterm: public Term {

	int m_edeg;
	Term   &term() {return *this;}
	static int validEdeg(int ed) {
		PRECONDITION(ed>=0);
		return ed;
	}
public:
	Eterm() {}
	Eterm(const Eterm &et):Term(et),m_edeg(et.edegree()) {}
	Eterm(const PolyContext &ctx):Term(ctx),m_edeg(0) {}
	Eterm(const PolyContext &ctx,int ed):Term(ctx),m_edeg(validEdeg(ed)) {}
	Eterm(const Term &t):Term(t),m_edeg(0) {}
	Eterm(const Term &t, int ed):Term(t),m_edeg(validEdeg(ed)) {}
	template<class Iter>
	Eterm(const PolyContext &ctx, int ed, Iter iter):Term(ctx,iter), m_edeg(validEdeg(ed)) {}

	Eterm &operator=(const Eterm &et) {
		Term::operator=(et);
		m_edeg=et.edegree();
		return *this;
	}

	const Term &term() const {return *this;}
	const Term &main() const {return term();}
	int edegree() const {return m_edeg;}

	bool operator==(const Eterm &et) const {
		return context().eto().equal(*this,et);
	}
	bool operator!=(const Eterm &et) const {
		return !context().eto().equal(*this,et);
	}

	bool operator<(const Eterm &et) const {
		return context().eto().less(*this,et);
	}
	bool operator>(const Eterm &et) const {
		return context().eto().less(et,*this);
	}

	int degree() const {
		return term().degree()+edegree();
	}

	int tdegree() const {
		return term().degree();
	}

	Eterm gcd(const Eterm &et) const {
		PRECONDITION(!isZero() && !et.isZero());
		return Eterm(term().gcd(et.term()),std::min(edegree(),et.edegree()));
	}
	Eterm lcm(const Eterm &et) const {
		PRECONDITION(!isZero() && !et.isZero());
		return Eterm(term().lcm(et.term()),std::max(edegree(),et.edegree()));
	}

	bool isCoprime(const Eterm &et) const {
		return (edegree()==0 || et.edegree()==0) && term().isCoprime(et.term());
	}

	bool divides(const Eterm &et) const {
		PRECONDITION(!isZero());
		return edegree()<=et.edegree() && term().divides(et.term());
	}
	Eterm divide(const Eterm &et) const {
		PRECONDITION(!et.isZero());
		PRECONDITION(edegree()>=et.edegree());
		return Eterm(term().divide(et.term()),edegree()-et.edegree());
	}
	bool isUnit() const {
		return edegree()==0 && term().isUnit();
	}
	bool isZero() const {return context().eb()<=edegree();}
	Term unit() const {return Eterm(context());}


	Term var(int varInd) const;
	Term varPow(int varInd, int pow) const;

	Eterm operator*(const Term &t) const {
		return Eterm(term()*t,edegree());
	}
	Eterm &operator*=(const Term &t) {
		term()*=t;
		return *this;
	}

	Eterm operator*(const Eterm &t) const {
		return Eterm(term()*t,edegree()+t.edegree());

	}
	Eterm &operator*=(const Eterm &t) {
		term()*=t;
		m_edeg+=t.edegree();
		return *this;
	}

	void setEdegree(int ed) {
		PRECONDITION(edegree()<=ed);
		m_edeg=ed;
	}
	void reduceEdegree(int ed) {
		PRECONDITION(ed<edegree());
		m_edeg=ed;
	}
	void resetEdegree() {
		m_edeg=0;
	}
	void print(std::ostream &os) const;
};

inline std::ostream &operator<<(std::ostream &os, const Eterm &et) {
	et.print(os);
	return os;
}



#endif //__ETERM_H__
