#include <cctype>
#include <iostream>
#include "checks.h"
#include "fpoint.h"

using namespace std;

const char * const Fpoint_error::prefix="MP floating point error: ";

Fpoint Fpoint::pow(long k) const {
	Fpoint r;
	if (k>0) {
		mpf_pow_ui(r.rep, rep, k);
	}
	else {
		PRECONDITION(!isZero());
		mpf_pow_ui(r.rep, (Fpoint(1)/(*this)).rep, -k);
	}		
	return r;
}

void Fpoint::write(ostream& os, const mpf_t src) {
  int base=10;
  const char *base_prefix="";
  switch (os.flags() & ios::basefield) {
    case ios::hex:
                base=16;
                base_prefix="0x";
                break;
    case ios::oct:
                base=8;
                base_prefix="0";
                break;
    default:;
  }
  mp_exp_t exp;
  char *mantissa=outbuf();
  char *t=mpf_get_str(mantissa,&exp,base,digits(),src);
  if (os.flags() & ios::showbase) os << base_prefix;
  if (*t == '-') os.put(*t++);
  if (exp<-4 || exp>5) {
    os<<*t++;
    if (*t) os<<'.'<<t;
    --exp;
    os<<'e';
    if (exp>0) os<<'+';
    if (exp<0) {os<<'-';exp=-exp;}
    if (exp<10) os<<'0';
    os<<exp;
  }
  else
  if (exp<=0) {
    os<<'0';
    if (*t) {
      os<<'.';
      while(++exp<=0) {os<<'0';}
      os<<t;
    }
  }
  else {
    do {
      if (*t) os<<*t++;else os<<'0';
    } while (--exp>0);
    if (*t) os<<'.'<<t;
  }  
}

ostream &Fpoint::print(ostream &os) const {
  write(os,rep);
  return os;
}

class Parse_fpoint {
  string text;
  char c;
  istream &is;
  
  void is_bad_eos() const;
  void skip_space();
  void sign();
  void mant();
  void exp();
  void signed_int();
  void signed_float();
public:

  Parse_fpoint(istream &i):is(i) {}
  
  string &operator()();
    
};

void Parse_fpoint::is_bad_eos() const {
  if (is.eof()) throw Fpoint_error("unexpexted end of read stream");
}

void Parse_fpoint::skip_space() {
  is>>c;
}

void Parse_fpoint::sign() {
  switch (c) {
	case '-':
	  text+=c;
	 case '+':
	  is.get(c);
  }
}

void Parse_fpoint::signed_int() {
  sign();
  is_bad_eos();
  while(isdigit(c)) {
	text+=c;
	is.get(c);
	if (is.eof()) break;
  }
}

void Parse_fpoint::signed_float() {
  sign();
  is_bad_eos();
  int dots=0;
  for(;;) {
	if (c=='.') {
	  if (++dots>1) throw Fpoint_error("too many dots");
	}
	else	if (!isdigit(c)) break;
	text+=c;
	is.get(c);
	if (is.eof()) break;
  }
}

void Parse_fpoint::mant() {
  signed_float();
}

void Parse_fpoint::exp() {
  if (c=='e' || c=='E') {
    text+=c;
    is.get(c);
    is_bad_eos();
    signed_int();
  }
};

string &Parse_fpoint::operator()() {
  skip_space();
  mant();
  exp();
  if (is && !is.eof()) is.putback(c);
  return text;
}

void Fpoint::read(istream& is, mpf_t dst)  noexcept(false) {
  Parse_fpoint parser(is);
  const int base=10;
  if ( mpf_set_str(dst,parser().c_str(),base) ) throw Fpoint_error("bad read format");
}

istream &Fpoint::scan(istream &is) noexcept(false) {
  try {
    read(is,rep);
  }
  catch(Fpoint_error &) {
    is.setstate(ios::failbit);
    throw;
  }
  return is;
}

struct Fpoint::Precision Fpoint::default_precision;
