#if !defined(__INPUTPARSER_H__)
#define __INPUTPARSER_H__

#include <string>
#include <map>
#include "polinparser.h"
#include "untyped.h"


class InputParser: public ParserBase {

	QueueScanner m_scanner;
	Error m_error;

public:

	struct Polins: public list<Polin> {
		void print(ostream &os) const;
	};

  struct Results: public std::map<std::string,Untyped>	{
		Untyped &operator[](const std::string &key) {
			return std::map<std::string,Untyped>::operator[](key);
		}
		const Untyped &operator[](const std::string &key) const {
			const_iterator pos=find(key);
			PRECONDITION(pos!=end());
			return pos->second;
		}
	};

private:
	Results m_results;

	bool have(const string &str) const {return m_results.find(str)!=m_results.end();}
public:
	InputParser():ParserBase(m_scanner,m_error) {}
	InputParser(  std::istream &is ):ParserBase(m_scanner,m_error),m_scanner(is) {}
	void load(std::istream &is) {m_scanner.load(is);}
	void reset() {m_scanner.reset();}
	bool operator()();

	const Results &results() const {return m_results;}
};

inline std::ostream &operator<<(std::ostream &os,const InputParser::Polins &pols) {
	pols.print(os);
	return os;
}

#endif //__INPUTPARSER_H__

