#if !defined(__LRAND48_H__)
#define __LRAND48_H__

#include <random>

struct random_factory {
  using Engine_I = std::uniform_int_distribution<std::uint_fast64_t>;
  using Engine_R = std::uniform_real_distribution<>;

  public:
  using type_i = Engine_I::result_type;
  using type_r = Engine_R::result_type;

  public:
  random_factory():m_gen(m_rd()),m_rand_i(),m_rand_r() {}

  public:
  type_i lrand48() { return m_rand_i(m_gen);}
  type_r drand48() { return m_rand_r(m_gen);}
  void  srand48(int s) {m_gen.seed(s);}

  static random_factory &instance() {
    static random_factory s_instance;
    return s_instance;
  }

  private:
  std::random_device m_rd;  //Will be used to obtain a seed for the random number engine
  std::mt19937 m_gen; //Standard engine seeded with rd()
  Engine_I m_rand_i;
  Engine_R m_rand_r;
};

inline void srand48(int s) {
  random_factory::instance().srand48(s);
}

inline long int lrand48() {
  return static_cast<long int>(random_factory::instance().lrand48());
}

inline double drand48() {
  return static_cast<double>(random_factory::instance().drand48());
}


#endif // __LRAND48_H__
