#include "monom.h"

Monom Monom::operator*(const Monom &m) const {
	return Monom(coeff()*m.coeff(),eterm()*m.eterm());
}

Monom &Monom::operator*=(const Monom &m) {
	coeff()*=m.coeff();
	eterm()*=m.eterm();
	return *this;
}

Monom Monom::operator*(const Coeff &c) const {
	return Monom(coeff()*c,m_eterm);
}

Monom &Monom::operator*=(const Coeff &c) {
	coeff()*=c;
	return *this;
}

Monom Monom::operator*(const Eterm &et) const {
	return Monom(m_coeff,eterm()*et);
}

Monom &Monom::operator*=(const Eterm &et) {
	eterm()*=et;
	return *this;
}

Monom Monom::operator/(const Coeff &c) const {
	PRECONDITION(!c.isZero());
	return Monom(coeff()/c,m_eterm);
}

Monom &Monom::operator/=(const Coeff &c) {
	PRECONDITION(!c.isZero());
	coeff()/=c;
	return *this;
}

Monom Monom::operator+(const Monom &m) const {
	PRECONDITION(eterm()==m.eterm());
	if ( edegree()==m.edegree() )
		return Monom(coeff()+m.coeff(),m_eterm);
	else
		return Monom(coeff()+m.coeff(),Eterm(eterm().term(),std::max(edegree(),m.edegree())));
}

Monom &Monom::operator+=(const Monom &m) {
	PRECONDITION(eterm()==m.eterm());
	coeff()+=m.coeff();
	eterm().setEdegree(std::max(edegree(),m.edegree()));
	return *this;
}

Monom Monom::operator-(const Monom &m) const {
	PRECONDITION(eterm()==m.eterm());
	if ( edegree()==m.edegree() )
		return Monom(coeff()-m.coeff(),m_eterm);
	else
		return Monom(coeff()-m.coeff(),Eterm(eterm().term(),std::max(edegree(),m.edegree())));
}

Monom &Monom::operator-=(const Monom &m) {
	PRECONDITION(eterm()==m.eterm());
	coeff()-=m.coeff();
	eterm().setEdegree(std::max(edegree(),m.edegree()));
	return *this;
}

void Monom::print(std::ostream &os) const {
	bool pt=true;
	if ( coeff()==-coeff().unit() ) os<<'-';
	else
	if ( !coeff().isUnit() ) {
		os<<coeff();
		if (eterm().isUnit()) pt=false;
		else os<<'*';
	}
	if (pt) os<<eterm();
}
