#if !defined __POLCONTEXT_H__
#define __POLCONTEXT_H__

#include <iostream>
#include "checks.h"
#include "alphabet.h"
#include "coeff.h"
#include "termorder.h"
#include "etermorder.h"
#include "shared_object.h"

class PolyContext {
	struct Context {
		Alphabet m_abc;
		TermOrder m_to;		
		EtermOrder m_eto;
		Coeff m_coeff;
		AlmostEqual m_almostEqual;
		int m_eb;
		Context( 	const Alphabet &a_abc,
						 	const TermOrder &a_to,
							const EtermOrder &a_eto,
							const Coeff &a_coeff,
							const AlmostEqual &ae,
							int a_eb):m_abc(a_abc),
												m_to(a_to),
												m_eto(a_eto),
												m_coeff(a_coeff),
												m_almostEqual(ae),
												m_eb(a_eb) {
			PRECONDITION(m_eb>=0);
			PRECONDITION(m_coeff.type()==m_almostEqual.epsilon().type());
		}
	};

	shared_object<Context> data;

public:

	PolyContext(  const Alphabet &a_abc,
								const TermOrder &a_to,
								const EtermOrder &a_eto,
								const Coeff &a_coeff,
								const AlmostEqual &ae,
								int a_eb):data(Context(a_abc,a_to,a_eto,a_coeff,ae,a_eb)) {}
	const Alphabet &abc() const {return data->m_abc;}
	const TermOrder &to() const {return data->m_to;}
	const EtermOrder &eto() const {return data->m_eto;}
	const Coeff &coeff() const {return data->m_coeff;}
	const AlmostEqual &almostEqual() const {return data->m_almostEqual;}
	int eb() const {return data->m_eb;}

	bool operator==(const PolyContext &po) const {
		return &*data==&*po.data;
	}

	bool operator!=(const PolyContext &po) const {
		return &*data!=&*po.data;
	}

	void print(std::ostream &os) const;

	static PolyContext defaultContext;
};

inline std::ostream &operator<<(std::ostream &os, const PolyContext &pc) {
	pc.print(os);
	return os;
}
#endif //__POLCONTEXT_H__
