#if !defined __POLINPARSER_H__
#define __POLINPARSER_H__

#include "polin.h"
#include "baseparser.h"

class IntPolinParser: public ParserBase {

	PolyContext m_context;
	bool m_exact;
	Polin m_zero;
	Polin m_unit;

	bool atom(Polin &);
	bool exp(Polin &);
	bool mul(Polin &);
	bool sum(Polin &val);

	bool exact() const {return m_exact;}
	Polin zero() const {return m_zero;}
	Polin unit() const {return m_unit;}
public:
	IntPolinParser(  QueueScanner &scanner,
									 Error &err,
									 const PolyContext &ctx
							  ):ParserBase( scanner,err ),
									m_context(ctx),
									m_exact(ctx.coeff().exact()),
									m_zero( Polin() ),
									m_unit( Polin(Monom(ctx.coeff().unit(),Term(ctx).unit())) )  {}

	bool expr(Polin &val) {
		val=zero();
		return sum(val);
	}

	const PolyContext &context() const {return m_context;}
};


class PolinParser: public IntPolinParser {

	QueueScanner m_scanner;
	Error m_error;
public:
	PolinParser( const PolyContext &ctx):IntPolinParser(m_scanner,m_error,ctx) {}
	PolinParser(  std::istream &is,
								const PolyContext &ctx
						 ):IntPolinParser(m_scanner,m_error,ctx),m_scanner(is) {}
	void load(std::istream &is) {m_scanner.load(is);}
	void reset() {m_scanner.reset();}
};


#endif //__POLINPARSER_H__
