#if !defined __PRINT2D_H__
#define __PRINT2D_H__

#include <string>
#include <sstream>
#include <ostream>
#include <iterator>
#include <map>
#include <algorithm>
#include <function.h>
#include <iomanip>

template< class __Elem_matter,
          class __Cont2D,
          class __Sub_cont = typename std::iterator_traits<typename __Cont2D::const_iterator>::value_type,
          class __Elem_type = typename std::iterator_traits<typename __Sub_cont::const_iterator>::value_type>
class Print2D {
  
public:  
  typedef __Cont2D Container;
  typedef __Sub_cont Sub_container;
  typedef __Elem_type Elem_type;
  typedef __Elem_matter Elem_matter;
  
  typedef typename Container::const_iterator row_const_iterator;
  typedef typename Container::iterator row_iterator;

  typedef typename Sub_container::const_iterator col_const_iterator;
  typedef typename Sub_container::iterator col_iterator;
  
private:
  
  struct Comp_rows {
	Comp_rows(const Print2D  &_par):par(_par) {}
	bool operator()(row_const_iterator x,  row_const_iterator y) const { return par.comp_rows(x,y);}
  protected:
	const Print2D &par;
  };

  struct Comp_cols {
	Comp_cols(const Print2D  &_par):par(_par) {}
	bool operator()(col_const_iterator x,  col_const_iterator y) const { return par.comp_cols(x,y);}
  protected:
	const Print2D &par;
  };

  friend class Comp_rows;
  friend class Comp_cols;
  
  virtual bool comp_rows(row_const_iterator x,  row_const_iterator y) const=0;
  virtual bool comp_cols(col_const_iterator x,  col_const_iterator y) const=0;
  
  virtual std::string r_name(const Container &c, row_const_iterator i) const {
		std::ostringstream r;
		r<<distance(c.begin(),i);
		return r.str();
  }
  
  virtual std::string c_name(const Sub_container &s, col_const_iterator j) const {
		std::ostringstream r;
		r<<distance(s.begin(),j);
		return r.str();
  }
  
  virtual const Sub_container &sub_container(row_const_iterator i) const=0;
  
  virtual const Elem_matter &elem_matter(const Elem_type &elem) const=0;
  
public:

  Print2D() {}
	virtual ~Print2D() {}
      
  std::ostream &operator()(std::ostream &os,const Container &c,int max_cols = 0);
  
protected:
    
  typedef std::map<std::string,std::string> Scols;
  typedef std::map<row_const_iterator,Scols,Comp_rows> Srows;
  
  typedef Srows Matrix;

  typedef std::map<row_const_iterator,std::string,Comp_rows> R_names;
  typedef std::map<col_const_iterator,std::string,Comp_cols> C_names;
  typedef typename R_names::const_iterator r_names_const_iterator;
  typedef typename C_names::const_iterator c_names_const_iterator;

  bool no_col( const C_names &ci, typename C_names::data_type offs ) {
	return ( ci.end()==find_if(ci.begin(),ci.end(),
				 compose1(
					 bind2nd(
						 equal_to<typename C_names::data_type>(),offs
					 ),
					 select2nd<typename C_names::value_type  >() )
				 ) 
		  );	
  }  
};

template<class __Elem_matter,class __Cont2D, class __Sub_cont, class __Elem_type>
std::ostream &Print2D<__Elem_matter,__Cont2D,__Sub_cont,__Elem_type>::operator()(std::ostream &os,const Container &c,int max_cols) {

  Comp_rows comp_rows(*this);
  Comp_cols comp_cols(*this);
  
  R_names r_names(comp_rows);

  Srows m(comp_rows);
  C_names c_names(comp_cols);
  
  std::map<std::string,int> tabs;
  int tab0=0;
  
  for (row_const_iterator i=c.begin();i!=c.end();++i) {
    std::string rname=r_name(c,i);
    int rname_size=rname.size();
    if (tab0<rname_size) tab0=rname_size;
    r_names[i]=rname;
    const Sub_container &sub=sub_container(i);
    for (col_const_iterator j=sub.begin();j!=sub.end();++j) {
      std::string cname=c_name(sub,j);
      if (no_col(c_names,cname)) {
        c_names[j]=cname;
        int cname_size=cname.size();
        tabs[cname]=cname_size;
      }	   
      std::string val;
      val<<elem_matter(*j);
      int val_size=val.size();
      if (tabs[cname]<val_size) tabs[cname]=val_size;
      m[i][cname]=val;
    }
  }
  
  os.flags(std::ios::right);
  
  typedef typename C_names::const_iterator c_names_const_iterator;
  c_names_const_iterator a=c_names.begin();
  c_names_const_iterator b=c_names.begin();
  
  if (a==c_names.end()) os<<"<empty>\n";
  else
  while(a!=c_names.end()) {  
    fill_n(std::ostream_iterator<char>(os),tab0+1,' ');
    os<<'|';
    int row_size=0;
    while( b!=c_names.end() ) {
      std::string cname=b->second;
      int tab=tabs[cname];
      row_size+=tab+1;
      os<<setw(tab+1)<<cname.c_str();
      ++b;
      if (max_cols>0 && row_size+tab0+2>max_cols) break;
    }
    os<<'\n';
    fill_n(std::ostream_iterator<char>(os),tab0+1,'-');
    os<<'+';
    fill_n(std::ostream_iterator<char>(os),row_size,'-');
    os<<'\n';
    for (typename Srows::const_iterator i=m.begin();i!=m.end();++i) {
      os<<setw(tab0+1)<<r_names[i->first].c_str()<<'|';
      for(c_names_const_iterator j=a;j!=b;++j) {
        const std::string &cname=j->second;
        const std::string &val=m[i->first][cname];
        int tab=tabs[cname];
        if (val.empty()) {
          fill_n(std::ostream_iterator<char>(os),tab,' ');
          os<<'.';	  
        }
        else
        os<<setw(tab+1)<<val.c_str();
      }
      os<<'\n';
    }
    a=b;
  }
  return os;
}

#endif // __PRINT2D_H__
