#if !defined(__PRINTEMATRIX_H__)
#define __PRINTEMATRIX_H__

#include <functional>
#include <set>
#include <map>
#include "checks.h"
#include "ematrix.h"

class Ptable {
	typedef unsigned int index_type;
	struct Tab: public std::map<Eterm,index_type,std::greater<Eterm> > {
		typedef std::map<Eterm,index_type,std::greater<Eterm> > Inherited;
		typedef Inherited::mapped_type mapped_type;
		typedef mapped_type data_type;
		index_type operator[](const Eterm &t) const {
			const_iterator pos=find(t);
			PRECONDITION(pos!=end());
			return pos->second;
		}
		index_type &operator[](const Eterm &t) {
			iterator pos=find(t);
			PRECONDITION(pos!=end());
			return pos->second;
		}
	};
	Tab m_tab;
public:
	Ptable(const SetOfEterms &st);
	bool isBefore(const Eterm &et1, const Eterm &et2) const {
		return m_tab[et1]<m_tab[et2];
	}
	index_type operator[](const Eterm &et) const {
		return m_tab[et];
	}
	void swap(const Eterm &et1,const Eterm &et2) {
		if ( isBefore(et1,et2) ) std::swap(m_tab[et1],m_tab[et2]);
	}
	bool contains(const Eterm &et) const {
		return m_tab.find(et)!=m_tab.end();
	}
	bool empty() const {return m_tab.empty();}
	const Eterm &maximal() const;
};

class ColLess: public std::binary_function<Eterm,Eterm,bool> {
	linked_object<Ptable> m_tab;
public:
	ColLess(const SetOfEterms &st):m_tab(st) {}
	bool operator()(const Eterm &et1,const Eterm &et2) const;
	void swap(const Eterm &et1,const Eterm &et2) {
		m_tab->swap(et1,et2);
	}
	const Ptable &ptable() const {return *m_tab;}
};

class Row;

class RowLess: public std::binary_function<Eterm,Eterm,bool> {
	ColLess m_colComp;
public:	
	RowLess(const ColLess &colComp):m_colComp(colComp) {}
	bool operator()(const Row &r1,const Row &r2) const;
	void swap(const Eterm &et1,const Eterm &et2) {
		m_colComp.swap(et1,et2);
	}
	ColLess colLess() const {return m_colComp;}
};

class Rows;

class Tabs: public std::map<Eterm,int,ColLess> {
	typedef std::map<Eterm,int,ColLess> Inherited;
	typedef Inherited::key_type key_type;
	typedef Inherited::mapped_type mapped_type;
	typedef mapped_type data_type;
public:
	data_type &operator[](const key_type &key) {
		iterator q=find(key);
		if (q!=end()) return q->second;
		return insert(q,value_type(key,0))->second;
	}
	data_type operator[](const key_type &key) const {
		const_iterator q=find(key);
		PRECONDITION(q!=end());
		return q->second;
	}
	Tabs(const Rows &rows);
	void selfAdjust();
	const_iterator firstIrred() const;
};

class Row: public std::map<Eterm,std::string,ColLess> {
	typedef std::map<Eterm,std::string,ColLess> Inherited;
	std::string m_tag;
public:
	Row(const Tpolin &bp, const ColLess &cl);
	Row(Tabs &);
	void tuneTabs(int &tagLen,Tabs &) const;
	const std::string &tag() const {return m_tag;}
};

std::ostream &operator<<(std::ostream &os,const Row &row);

class Rows: public std::multiset<Row,RowLess> {
	typedef std::multiset<Row,RowLess> Inherited;
public:
	Rows(const SetOfEterms &st):Inherited(RowLess(ColLess(st))) {}
};

class PrintEmatrix {
	Rows m_rows;
	Tabs m_tabs;
	static std::pair<Tabs::const_iterator,int> printRow(std::ostream &os, Tabs::const_iterator watershed, const Row &row, Tabs::const_iterator first,Tabs::const_iterator last,int pageWidth);
public:
	PrintEmatrix(const SetOfEterms &st):m_rows(st),m_tabs(m_rows) {}
	std::ostream &operator()(std::ostream &os, const Elimination::Ematrix &em ,int pageWidth=1000000);
	void swap(const Eterm &et1,const Eterm &et2) const {
		//if ( m_rows.key_comp().colLess().ptable().contains(et1) )
		m_rows.key_comp().swap(et1,et2);
	}
};


#endif //__PRINTEMATRIX_H__
