#if !defined(__REPORT_H__)
#define __REPORT_H__

#include <fstream>
#include <string>
#include <map>
#include "checks.h"
#include "shared_object.h"

class Report {

	struct Data {
		std::string filename;
		std::ofstream os;
		int printLevel;
		int matrWidth;
		mutable int ic;

		Data(const char *fname,int pl,int mw):filename(fname),os(fname),printLevel(pl),matrWidth(mw),ic(1) {}

		void lock() const {
			PRECONDITION(ic>0);
			++ic;
		}
		void release() const {
			PRECONDITION(ic>0);
			if (--ic==0) delete this;
		}
		void print(std::ostream &os) const;
	};

	Data *m_data;	

public:
	
	Report(const char *fname, int pl=1,int mw=180);
	Report(const Report &r):m_data(r.m_data) {
		m_data->lock();
	}
	
	Report &operator=(const Report &r) {
		r.m_data->lock();
		m_data->release();
		m_data=r.m_data;
		return *this;
	}

	~Report() {m_data->release();}
	int printLevel() const {return m_data->printLevel;}
	void setPrintLevel(int pl) {m_data->printLevel=pl;}
	int matrWidth() const {return m_data->matrWidth;}
	void setMatrWidth(int mw) {m_data->matrWidth=mw;}
	std::ostream &rs() {return m_data->os;}
	const std::string &fileName() const {return m_data->filename;}

	Report &report() {return *this;}
	const Report &report() const {return *this;}
	void print(std::ostream &os) const {
		m_data->print(os);
	}
};

inline std::ostream &operator<<(std::ostream &os, const Report &r) {
	r.print(os);
	return os;
}

#define REPORT(lev,data) if (printLevel()>=lev) {rs()<<data;}




class Stat  {

	class Data {
		typedef std::map<std::string,int> Vals;
		typedef std::map<std::string,std::string> Descr;
		Vals m_vals;
		Descr m_descr;
	public:
		bool add(const std::string &key,const std::string &comment);
		bool add(const std::string &key) {
			return add(key,key);
		}
		int &operator[](const std::string &key);
		void print(std::ostream &os) const;
		bool empty() const {return m_vals.empty();}
	};

	linked_object<Data> m_data;

public:
	Stat():m_data(Data()) {}
	Data &stat() {return *m_data;}
	const Data &stat() const {return *m_data;}
};


inline std::ostream &operator<<(std::ostream &os, const Stat &s) {
	s.stat().print(os);
	return os;
}

#endif //__REPORT_H__
