#if !defined(__SETOFTERMS_H__)
#define __SETOFTERMS_H__

#include <set>
#include <algorithm>
#include <functional>
#include <ostream>
#include "checks.h"
#include "term.h"

// An adapted binary_function originated from:
//  Adapters and Binders - Overcoming problems
//  in the design and implementation of the C++-STL
//  Volker Simonis
//  Universitat Tubingen


namespace std {
  template <class Arg1, class Arg2, class Result>
  struct binary_function<Arg1, Arg2&, Result> {
    typedef Arg1 first_argument_type;
    typedef Arg2 second_argument_type;
    typedef Result result_type;
  };

  template <class Arg1, class Arg2, class Result>
  struct binary_function<Arg1, const Arg2&, Result> {
    typedef Arg1 first_argument_type;
    typedef Arg2 second_argument_type;
    typedef Result result_type;
  };
}
/////////////////////////////////////////////////////////////


class SetOfTerms: public std::set<Term> {


public:
	void print(std::ostream &os) const;
	bool isZeroDim() const;
	bool contains(const Term &t) const {
		return find(t)!=end();
	}
	SetOfTerms nset() const;
	SetOfTerms bset() const;
};

template<class ST,class T2term>
inline bool isZeroDim(const ST &st, T2term t2term) {
	if (st.empty()) return true;
	std::set<int> pps;
	size_t n = t2term(*st.begin()).context().abc().size();
	for (typename ST::const_iterator i=st.begin();i!=st.end();++i) {
		int var=-1;
		if (t2term(*i).isPP(var)) {
			pps.insert(var);
			if (pps.size()==n) break;
		}
	}
	return pps.size()==n;
}

template<class ST,class T2term>
inline SetOfTerms nset(const ST &st, T2term t2term) {
	PRECONDITION(isZeroDim(st,t2term));
	SetOfTerms ns;
	if (st.empty()) return ns;
	SetOfTerms band,next;
	PolyContext ctx=t2term(*st.begin()).context();
	Term unit=Term(ctx).unit();
	band.insert(unit);	
	int n=ctx.abc().size();
	while (!band.empty()) {
		for (SetOfTerms::const_iterator t=band.begin();t!=band.end();++t) {
			for (int i=0;i<n;++i) {
				Term s=*t*unit.var(i);
				if (std::find_if(st.begin(),st.end(), __gnu_cxx::compose1(std::bind2nd(std::mem_fun_ref(&Term::divides),s),t2term)) == st.end()) next.insert(s);
			}
		}
		ns.insert(band.begin(),band.end());
		band=next;
		next.clear();
	}
	return ns;
}

inline std::ostream &operator<<(std::ostream &os, const SetOfTerms &st) {
	st.print(os);
	return os;
}


#endif //__SETOFTERMS_H__

