#include <numeric>
#include <algorithm>
#include <functional>
#include <iostream>
#include "term.h"

using namespace std;

bool Term::isUnit() const {
	return std::find_if( begin(),end(),std::bind2nd(std::not_equal_to<int>(),0) )==end();
}

void Term::print(std::ostream &os) const {
	if (isUnit()) os<<'1';
	else {
		const_iterator e=std::find_if( rbegin(),rend(),std::bind2nd(std::not_equal_to<int>(),0) ).base();
		PRECONDITION(begin()!=e);--e;
		for(const_iterator i=begin();;++i) {
			int p = *i;
			if (p>0) os<<context().abc()[std::distance(begin(),i)];
			if (p>1) {
				os<<'^';
				//bool brace=(p>9);
				//if (brace) os<<'{';
				os<<p;
				//if (brace) os<<'}';			
			}
			if (i==e) break;
			if (p>0) os<<'*';
		}
	}
}

Term Term::operator*(const Term &t) const {
	sameContext(t);
	Term result(context());
	std::transform(begin(),end(),t.begin(),result.begin(),std::plus<int>());
	return result;
}

Term &Term::operator*=(const Term &t) {
	sameContext(t);
	std::transform(begin(),end(),t.begin(),begin(),std::plus<int>());
	return *this;
}

Term Term::var(int varInd) const {
	PRECONDITION(varInd>=0);
	Term result=unit();
	result[varInd]=1;
	return result;
}

Term Term::varPow(int varInd, int pow) const {
	PRECONDITION(varInd>=0 && pow>0);
	Term result=unit();
	result[varInd]=pow;
	return result;
}

Term Term::gcd(const Term &t) const {
	Term result(context());
	std::transform(begin(),end(),t.begin(),result.begin(),ptr_fun<const int &,const int &,const int &>(std::min<int>) );
	return result;	
}

Term Term::lcm(const Term &t) const {
	Term result(context());
	std::transform(begin(),end(),t.begin(),result.begin(),ptr_fun<const int &,const int &,const int &>(std::max<int>) );
	return result;
}

bool Term::isCoprime(const Term &t) const {
	sameContext(t);
	const_iterator i1=begin();
	const_iterator i2=t.begin();
	while (i1!=end()) {
		if (*i1!=0 && *i2!=0) return false;
		++i1;++i2;
	}
	return true;
}

bool Term::isPP(int &var) const {
	if (count_if(begin(),end(),bind2nd(not_equal_to<int>(),0))==1) {
		const_iterator i=find_if(begin(),end(),bind2nd(not_equal_to<int>(),0));
		PRECONDITION(i!=end());
		var=distance(begin(),i);
		return true;
	}
	return false;
}

bool Term::divides(const Term &t) const {
	sameContext(t);
	const_iterator i1=begin();
	const_iterator i2=t.begin();
	while (i1!=end()) {
		if (*i2<*i1) return false;
		++i1;++i2;
	}
	return true;
}

Term Term::divide(const Term &t) const {
	sameContext(t);
	PRECONDITION(t.divides(*this));
	Term result(context());
	transform(begin(),end(),t.begin(),result.begin(),minus<int>());
	return result;
}
