#include <iostream>
#include <algorithm>
#include "termorder.h"
#include "term.h"

using std::ostream;

bool VectorOrder::equal(const Term &t1, const Term &t2) const {
  if (t1.begin()==t2.begin()) return true;
	return std::equal( t1.begin(),t1.end(),t2.begin() );
}

struct Lex: public VectorOrder {
	virtual bool less(const Term &t1, const Term &t2) const;
	virtual void print(ostream &os) const;	
};

bool Lex::less(const Term &t1, const Term &t2) const {
  if (t1.begin()==t2.begin()) return false;
	return std::lexicographical_compare( t1.begin(),t1.end(),t2.begin(),t2.end() );
}

void Lex::print(ostream &os) const {
	os<<"Pure lexicographical";
}

// RevLex is not an admissible order !

struct RevLex: public VectorOrder {
	virtual bool less(const Term &t1, const Term &t2) const;
	virtual void print(ostream &os) const;		
};

bool RevLex::less(const Term &t1, const Term &t2) const {
  if (t1.begin()==t2.begin()) return false;
	return std::lexicographical_compare(t2.rbegin(),t2.rend(),t1.rbegin(),t1.rend());
}

void RevLex::print(ostream &os) const {
	os<<"Reverse lexicographical";
}

struct Tdeg: public Lex {
	virtual bool less(const Term &t1, const Term &t2) const;
	virtual void print(ostream &os) const;		
};

bool Tdeg::less(const Term &t1, const Term &t2) const {
  if (t1.begin()==t2.begin()) return false;
	if (t1.degree()<t2.degree()) return true;
	if (t2.degree()<t1.degree()) return false;
	return Lex::less(t1,t2);
}

void Tdeg::print(ostream &os) const {
	os<<"Total degree";
}

struct RevTdeg: public RevLex {
	virtual bool less(const Term &t1, const Term &t2) const;
	virtual void print(ostream &os) const;		
};

bool RevTdeg::less(const Term &t1, const Term &t2) const {
  if (t1.begin()==t2.begin()) return false;
	if (t1.degree()<t2.degree()) return true;
	if (t2.degree()<t1.degree()) return false;
	return RevLex::less(t1,t2);
}

void RevTdeg::print(ostream &os) const {
	os<<"Reverse total degree";
}


TermOrder TermOrder::lex() {
	static Lex order;
	return TermOrder(&order);
}

TermOrder TermOrder::revlex() {
	static RevLex order;
	return TermOrder(&order);
}

TermOrder TermOrder::tdeg() {
	static Tdeg order;
	return TermOrder(&order);
}

TermOrder TermOrder::revtdeg() {
	static RevTdeg order;
	return TermOrder(&order);
}
